! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!

module gofutils_mod

  use dal
  use oal
  use errorHandling

  implicit none

  public

contains
  !**********************************************************************
  Pure Function to_upper (str) Result (string)

    !   ==============================
    !   Changes a string to upper case
    !   ==============================

    Implicit None
    Character(*), Intent(In) :: str
    Character(LEN(str))      :: string

    Integer :: ic, i

    Character(26), Parameter :: cap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    Character(26), Parameter :: low = 'abcdefghijklmnopqrstuvwxyz'

    !   Capitalize each letter if it is lowecase
    string = str
    do i = 1, LEN_TRIM(str)
       ic = INDEX(low, str(i:i))
       if (ic > 0) string(i:i) = cap(ic:ic)
    end do

  End Function to_upper

  !**********************************************************************
  Pure Function to_lower (str) Result (string)

    !   ==============================
    !   Changes a string to lower case
    !   ==============================

    Implicit None
    Character(*), Intent(In) :: str
    Character(LEN(str))      :: string

    Integer :: ic, i

    Character(26), Parameter :: cap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    Character(26), Parameter :: low = 'abcdefghijklmnopqrstuvwxyz'

    !   Capitalize each letter if it is lowecase
    string = str
    do i = 1, LEN_TRIM(str)
       ic = INDEX(cap, str(i:i))
       if (ic > 0) string(i:i) = low(ic:ic)
    end do

  End Function to_lower

  !**********************************************************************

  subroutine binit (time,rate,binsize,nchan,bin_rate,tstart,tend,n)
    !
    ! simple binning routine
    !
    integer :: i, nn
    integer, intent(in) :: binsize, nchan
    integer, intent(out) :: n
    real(kind=double), intent(in) :: time(nchan), rate(nchan)
    real(kind=double), intent(out) :: &
         bin_rate(nchan), tstart(nchan), tend(nchan)
    real(kind=double) :: t, c
    tstart = 0.d0
    tend = 0.d0
    t = time(1)
    c = rate(1)
    i = 2
    n = 1
    tstart(n) = time(1) - 0.5
    nn = 1
    do while (i .lt. nchan)
       do while (((time(i) - t) .lt. binsize) .and. (i .le. nchan-1))
          c = c + rate(i)
          nn = nn + 1
          i = i + 1
       enddo
       if(nn .gt. 0) then
          bin_rate(n) = c/nn
          tend(n) = time(i-1) + 0.5
          n = n + 1
       endif
       tstart(n) = time(i) - 0.5
       c = 0.0
       nn = 0
       t = time(i)
    end do
    return
  end subroutine binit
  !
  !
  !
  subroutine smooth(rate,nchan,smoo,smoo_rate)
    !
    ! simple smoothing routine
    !
    implicit none

    integer, intent(in) :: nchan, smoo
    integer :: i, ih, j, j1, j2
    real(kind=double) :: n, t = 0.
    real(kind=double), intent(in) :: rate(nchan)
    real(kind=double), intent(out) :: smoo_rate(nchan)
    smoo_rate = 0.d0
    ih = smoo/2
    do i=1,nchan
       j1 = i - ih
       if(j1 .lt. 1) j1 = 1
       j2 = i + ih
       if(j2 .gt. nchan) j2 = nchan
       t = 0.0
       n = 0.0
       do j=j1,j2
          t = t + rate(j)
          n = n + 1.0
       enddo
       smoo_rate(i) = t/n
    enddo
    return
  end subroutine smooth
  !
  ! BHP: used by: conv_reg_mod
  !
  subroutine read_region_ascii (infile,nRegions,shape,sh,&
       ra,dec,detx,dety,radius,rada,radb,rotang)
    integer :: i, lun, ios, status, nRegions, sh(nRegions)
    character(*) :: infile, shape(nRegions)
    character(len=1024) line ! BHP: check this!!!!
    real(kind=single) :: ra(nRegions), dec(nRegions), &
         detx(nRegions),dety(nRegions),radius(nRegions),&
         rada(nRegions),radb(nRegions),rotang(nRegions)

    status = 0

    lun = GetLun()

    open(unit=lun,status='old',form='formatted', &
         file=trim(infile),iostat=ios)
    if(ios .ne. 0) then
       call error ("noOpen","Unable to open input file")
    endif

    do i=1,nRegions
       read(lun,'(a)',iostat=ios) line
       if (ios /=0 ) return

       if ( (index(trim(line),'CIRCLE') .gt. 0 ) .or. &
            (index(trim(line),'circle') .gt. 0 )) then
       elseif ( (index(trim(line),'ELLIPSE') .gt. 0 ) .or. &
            (index(trim(line),'ellipse') .gt. 0 )) then

       end if
    end do
  end subroutine read_region_ascii

  !
  ! BHP: read a real32 table, return some parts
  !      used by: 
  !
  subroutine read_real_table(&
       infile,cdelt,submode,xSize,ySize,arrayDataReal32)

    implicit none
    type(DataSetT) :: inSet
    type(ArrayT) :: inArray
    character(len=*), intent(in) :: infile
    character(len=25), intent(out) :: submode
    integer, dimension(:), pointer :: dimVector
    integer, intent(out) :: xSize, ySize
    !    real(kind=single), intent(inout), dimension(:,:) :: image
    real(kind=single), dimension(:,:), pointer :: arrayDataReal32
    real(kind=double), intent(out) :: cdelt(2)

    inSet = dataSet(trim(infile), Read) 
    inArray = array(inSet, 0, READ)
    dimVector => dimensions(inArray)
    xSize = dimVector(0)
    ySize = dimVector(1)
    allocate(arrayDataReal32(xSize,ySize))
    if (dataType(inArray) .eq. REAL32) then
       arrayDataReal32 => real32Array2Data(inArray)
    else 
       call error("inArrayNOTreal32",&
            "read_real_table: datatype of input array not Real32")
    end if

    ! BHP: check CRVAL/CRPIX/BUNIT/etc !!!

    if (hasAttribute(inSet,'CDELT1')) then
       cdelt(1) = real64Attribute(inSet,'CDELT1')
    else
       call error("imgNOcdelt",&
            "read_real_table: input image has no CDELT")
    end if
    if(hasAttribute(inSet,'CDELT2')) then
       cdelt(2) = real64Attribute(inSet,'CDELT2')
    else
       call error("imgNOcdelt",&
            "read_real_table: input image has no CDELT")
    end if
    !
    if (hasAttribute(inSet,'SUBMODE')) then
       submode = stringAttribute(inSet,'SUBMODE')
    else
       submode = ""
    end if

    call release(inSet)
    return

  end subroutine read_real_table
  !
  ! BHP: new read_rmf, using SAS instead of FITSIO
  !      temporary, will replace with CAL_get...

  subroutine new_read_rmf (infile,inRows,rmfData,eminin,emaxin,&
       nchanrin,eminout,emaxout,nchanrout)

    implicit none

    type(DataSetT) :: inSet
    type(TableT) :: inTab
    type(ColumnT) :: eloCol, ehiCol, rmfCol, &
         ngrpCol, fchanCol, nchanCol
    character(len=*), intent(in) :: infile
    character(len=8) :: columns(6)
    integer :: i, inRows, nRows, nchanrin, nchanrout
    real(kind=single), dimension(:), pointer :: detm, & 
         rmfData, eloData, ehiData, ngrpData, fchanData, nchanData
    real(kind=single) :: eminin, emaxin, eminout, emaxout

    columns = (/ 'ENERG_LO', 'ENERG_HI', 'N_GRP   ', 'F_CHAN  ', &
         'N_CHAN  ', 'MATRIX  ' /)

    inSet = dataSet(trim(infile), READ) 

    if (.not.hasBlock(inSet,'MATRIX')) then  
       call error("noMATRIXinRMF",&
            "read_rmf: Input RMF file has no MATRIX extension")
    end if
    inTab = table(inSet,'MATRIX', READ)
    nRows = numberOfRows(inTab)
    if (nRows .ne. inRows) then  
       write(errstr,*) &
            "read_rmf: Input MATRIX table rows .ne. to ",inRows
       call error("nchanNEmatrix",trim(errstr))
    end if

    !  Check for RMF column data

    do i=1,6
       if (.not.hasColumn(inTab,trim(columns(i)))) then 
          write(errstr,*) &
               "read_rmf: Input RMF file has no ",&
               trim(columns(i))," column."
          call error("RMFnoCol",trim(errstr))
       end if
    end do

    eloCol = column(inTab,'ENERG_LO',READ)
    eloData => real32Data(eloCol)
    ehiCol = column(inTab,'ENERG_HI',READ)
    ehiData => real32Data(ehiCol)
    ngrpCol = column(inTab,'N_GRP',READ)
    ngrpData => real32Data(ngrpCol)
    fchanCol = column(inTab,'F_CHAN',READ)
    fchanData => real32Data(fchanCol)
    nchanCol = column(inTab,'N_CHAN',READ)
    nchanData => real32Data(nchanCol)
    rmfCol = column(inTab,'MATRIX',READ)
    rmfData => real32Data(rmfCol)

    call release(inTab)
    call release(inSet)
    return

  end subroutine new_read_rmf
  !
  ! BHP: temp subroutine to read new swcx files before
  !      CAL takes over
  !
  ! BHP: new read_rmf, using SAS instead of FITSIO
  !      temporary, will replace with CAL_get...
  !      SWCX CCF file structure:
  !      15 rows, 4 columns: DETECTOR, ELO, EHI, SWCXMAP

  subroutine read_swcx (infile,idet,line,swcxXsize,swcxYsize,swcxArray)

    implicit none

    type(DataSetT) :: inSet
    type(TableT) :: inTab
    type(ColumnT) :: inCol
    character(len=*), intent(in) :: infile
    character(len=8) :: lines(5)
    integer, intent(in) :: idet, swcxXsize, swcxYsize
    integer :: nRows, i, row
    logical :: found
    real(kind=single), intent(in) :: line
    real(kind=single), intent(out) :: swcxArray(swcxXsize,swcxYsize)
    real(kind=single) :: elo(5), ehi(5)
    real(kind=single), dimension(:,:), pointer :: swcxRowPtr
    !
    !   lines = &
    !        (/ 'M1OVII  ','M1OVIII ', 'M1OVIIIB', 'M1NEIX  ', 'M1MGXI  ',&
    !        'M2OVII  ','M2OVIII ', 'M2OVIIIB', 'M2NEIX  ', 'M2MGXI  ',&
    !        'PNOVII  ','PNOVIII ', 'PNOVIIIB', 'PNNEIX  ', 'PNMGXI  '/)
    elo = (/0.00, 0.56, 0.65, 0.81, 0.91/)
    ehi = (/0.56, 0.65, 0.81, 0.91, 1.34/)

    found = .FALSE.
    do i=1,5
       if ( (line .ge. elo(i)) .AND. (line .lt. ehi(i)) ) then
          found = .TRUE. 
          row = idet * 5 + i    ! BHP: M1=1,5; M2=6,10, PN=11,15
       end if
    end do

    write(errstr,*) "read_swcx: row in SWCX CCF = ",row
    call message(AppMsg,Always) ! BHP: change later!!!

    inSet = dataSet(trim(infile), READ) 
    if (.not.hasBlock(inSet,'SWCX')) then  
       call error("noSWCXinSWCX",&
            "read_swcx: Input SWCX file has no SWCX extension")
    end if
    inTab = table(inSet,'SWCX', READ)
    if (.not.hasColumn(inTab,'SWCXMAP')) then ! BHP: must name CCF col SWCXMAP!!!
       call error("noSWCXMAPinSWCX",&
            "read_swcx: Input SWCX table has no SWCXMAP column")
    end if
    inCol = column(inTab,'SWCXMAP',READ)
    !
    nRows = numberOfRows(inTab)
    if (nRows .ne. 15) then  
       write(errstr,*) &
            "read_swcx: Input SWCX table rows .ne. to 15"
       call error("nchanNEswcx",trim(errstr))
    end if

    !  Extrack correct SWCX row data for DET/LINE

    swcxRowPtr => real32Array2Data(inCol,row)

    !    allocate(swcxArray)

    swcxArray(1:swcxXsize,1:swcxYsize) = &
         swcxRowPtr(0:swcxXsize-1,0:swcxYsize-1)

    call release(inCol)
    call release(inTab)
    call release(inSet)
    return

  end subroutine read_swcx

  !
  ! BHP: write out ascii QDP (QuickDirtyPlotter) file
  !      used by: 
  !
  subroutine write_QDP_file(QDPfile,QDP_array,nlines)

    implicit none
    integer, intent(in) :: nlines
    integer :: i, lun
    character(len=*), intent(in) :: QDPfile
    character(len=80), intent(in) :: QDP_array(nlines)

    lun = GetLun()

    if ( clobber() ) then 
       open(unit=lun,status='replace',form='formatted',file=trim(QDPfile))
    else
       call error ("noClobber",&
            "write_QDP_file: File exists but no clobber set to TRUE.")
    endif

    do i=1,nlines
       write(lun,*) trim(QDP_array(i))
    end do

    close(lun)
    return

  end subroutine write_qdp_file

  integer function GetLun ()
    ! by Steve Lionel, Ron Shepherd, William Clodius, Clive Page,
    ! Jeff Drummond. and others as posed at comp.lang.fortran on
    ! 9 September 1997
    implicit none
    logical :: exs, opn
    integer :: i
    getlun = -1  ! returned if no units are available.
    do i=7,99
       inquire (unit=i,exist=exs,opened=opn)
       if (exs .and. .not. opn) then
          getlun = i
          return
       end if
    end do
    write (*,*) "There are no free Fortran logical units available."
    return
  end function getlun
  !  
  subroutine histo(inarray,n,rangearray,m,bucket)
    !
    ! simple histogram filler
    !
    implicit none
    real(kind=double), dimension(1:), intent(in) :: inarray ! new_modrate_lc
    integer, intent(in) :: n  ! size(new_modrate_lc)
    ! rangearray, e.g. range edges from clean_lc in espfilt
    real(kind=double), dimension(1:), intent(in) :: rangearray 
    integer, intent(in)                :: m       ! # of ranges
    integer                            :: i, j  
    integer, dimension(1:m+1)          :: bucket  ! counting bucket
    do i = 1, m+1                     ! clear buckets
       bucket(i) = 0
    end do
    do i = 1, n                       ! for each input score
       do j = 1, m                    ! determine the bucket
          if (inarray(i) < rangearray(j)) then
             bucket(j) = bucket(j) + 1
             exit
          end if
       end do                         ! don't forget the last bucket
       if (inarray(i) >= rangearray(m)) bucket(m+1) = bucket(m+1)+1
    end do
    
    return
    
  end subroutine histo  
  !  
  subroutine oldhisto(inarray,n,rangearray,m,bucket)
    !
    ! simple histogram filler
    !
    implicit none
    real(kind=double), dimension(:), intent(in) :: inarray ! new_modrate_lc
    integer, intent(in)                :: n  ! size(new_modrate_lc)
    real(kind=double), dimension(:), intent(in) :: rangearray ! range edges from clean_lc
    integer, intent(in)                :: m       ! # of ranges
    integer                            :: i, j  
    integer, dimension(1:m+1)          :: bucket  ! counting bucket
    do i = 1, m+1                     ! clear buckets
       bucket(i) = 0
    end do
    do i = 1, n                       ! for each input score
       do j = 1, m                    ! determine the bucket
          if (inarray(i) < rangearray(j)) then
             bucket(j) = bucket(j) + 1
             exit
          end if
       end do                         ! don't forget the last bucket
       if (inarray(i) >= rangearray(m)) bucket(m+1) = bucket(m+1)+1
    end do
    
    return
    
  end subroutine oldhisto  
  !
  !
  !
  subroutine maskchips (inst,ccd,arr_RP)
    ! routine to mask chips used in proton, swcx
    ! input: pointer array from proton, swcx
    ! output: new pointer array with mask pixels zeroed out
    !
    character(len=2), intent(in) :: inst
    integer, intent(in) :: ccd
    integer :: i1(7), i2(7), ii1(7), ii2(7), i, j
    real(single), intent(inout), dimension(:,:), pointer :: arr_RP

    if (inst .eq. 'M1')then 
       ! these are values for MOS1 as determined by KDK from XMM_SWCX_0001.CCF
       ! 05/10/22
       i1  = (/259,392,523,395,130,0,128/)
       i2  = (/520,651,779,652,392,258,391/)
       ii1 = (/258,0,252,520,520,257,0/)
       ii2 = (/517,249,514,779,779,519,252/)
       ! these are values for MOS2 as determined by KDK from XMM_SWCX_0001.CCF
       ! 05/10/22
    elseif (inst .eq. 'M2') then 
       i1  = (/258,391,525,391,126,0,125/)
       i2  = (/521,650,779,650,390,256,387/)
       ii1 = (/258,0,253,521,522,257,0/)
       ii2 = (/518,252,516,779,779,520,253/)
    elseif (inst .eq. 'PN') then
       i1   = (/23,346,346,23,0,0,0/)
       i2   = (/346,669,670,347,0,0,0/)
       ii1  = (/368,368,53,54,0,0,0/)
       ii2  = (/684,684,368,368,0,0,0/)
       if (ccd .gt. 4) then
          call error("FourQuads","There can only be four quads")
       endif
    else
       write(errstr,*)"Inst is not M1/M2/PN, it is: ",inst
       call error("BadInst",trim(errstr))
    endif
    ! these will need tweaked by KK/SS for SAS-19 ???
    ! BHP: 'old' masking updated above
    !i1mos  = (/259,392,524,394,130,1,129/)
    !i2mos  = (/522,652,780,654,393,258,391/)
    !ii1mos = (/259,1,253,521,321,257,1/)
    !ii2mos = (/522,251,516,780,780,520,253/)
    do i=i1(ccd),i2(ccd)
       do j=ii1(ccd),ii2(ccd)
          arr_RP(i,j) = 0.
       end do
    end do
    
    return

  end subroutine maskchips

  subroutine getootsca (submode,ootsca,warndie)                                         
    !
    ! more robust to handle any invocation. it was done seperately (and
    ! differently!) in several tasks
    ! Returns PN OOT scale factor based on SUBMODE
    !
    character(len=25),intent(inout) :: submode
    logical :: warndie, gotit
    real, intent(inout) :: ootsca
    
    ootsca = 0.063 ! by default ???!!!
    gotit  = .FALSE.
    submode = trim(adjustl(submode))
    if(submode(:15) .eq. 'PrimeFullWindow') then
       ootsca = 0.063
       if(submode(:23) == 'PrimeFullWindowExtended') then 
          ootsca = 0.0232
          if (submode(:24) == 'PrimeFullWindowExtended3') then
             ootsca = 0.0232
          elseif (submode(:24) == 'PrimeFullWindowExtended5') then
             ootsca = 0.0163
          end if
       end if
       gotit = .TRUE.
    end if
    if (.not.gotit) then
       ! Future: something specifically appropriate here later for
       ! FastBurst, FastTiming, PrimeLargeWindow, PrimeSmallWindow
       if (warndie) then ! if true from calling task, die gracefully here
          write(errstr,*)"Submode: ",trim(adjustl(submode)),&
               " not appropriate for ESAS analysis. Exiting"
          call error("BadSubmode",trim(errstr))
       else
          write(errstr,*)"Submode: ",trim(adjustl(submode)),&
               " not appropriate for ESAS analysis. May produce bad results"
          call warning("BadSubmode",trim(errstr))
       end if
    end if
    return
    
  end subroutine getootsca
  
end module gofutils_mod
